ASMB,A,B,L
*
* VDRIVE2 dos for embedding into HPBASIC 11/17/10 WTN
* also can save basic apps to PTP
* 9/24/10 - original "B" version
* 11/17/10 - "B2" version with driver fix
*
* Installation...
* Make sure USB EQU = your PTP or USB adapter slot
* As configured, works only for the premade basic1.abs binary
* Assemble udos_b2.asm using asm21tr.pl to produce an ABS binary
* Load basic1.abs into core
* Remove matrix instructions for more memory...
*  put 0 into locations 4110 and 4111
*  put 11456 into location 110
* Load udos_b2.abs into core
* Run from location 100
* Attach media to PTP or USB interface
* Enter BYE
* Select option 1 to save binary containing BASIC w/ UDOS
* If a USB interface, enter a filename otherwise just writes
*
* Note - when saving a binary to USB, if the file exists the
* new binary is appended to it. To replace, use the dos prompt
* option and use DEL filename to erase the file first. Appending
* the binary might be useful when creating different boot options,
* select next binary by pressing run at halt.
*
* To save a BASIC program as an app, enter the program, enter
* BYE then save. Upon reload start from location 2 to run the
* program (starting from location 100 erases the program).
* If running isn't desired start from location 2050 instead.
* 
       ORG 2B
       JMP *+1,I
       OCT 5137           2050 = re-enter, 5137 = start prog
       ORG 76B
       DEF ENTRY          BYE runs udos
       JMP *-1,I
       ORG 106B
       OCT 16347          loc of udos - 1
       ORG 116B
       OCT 0
TOMEM  EQU 17700B         Last memory location + 1 to save
TTY    EQU 11B            TTY console slot
USB    EQU 12B            USB interface slot
       ORG 16350B         location of udos
ENTRY  CLF 0
       CLC 0
       JMP MENU
* udos must end before 17217, start of BASIC's drivers
*
* HP-to-VDRIVE2 interface firmware commands...
* 120000 = 1010000000000000   read from VDRIVE, follow by read transfer
* 121bbb = 10100010bbbbbbbb   write to VDRIVE, byte in low 8 bits of command
* 122000 = 1010010000000000   sync VDRIVE
* 123000 = 1010011000000000   clear VDRIVE
* 130000 = 1011000000000000   open read file, follow by filename[cr]
* 131000 = 1011001000000000   read from file, follow by read transfer
* 132000 = 1011010000000000   read from file using paper-tape emulator mode
* 134000 = 1011100000000000   open write file, follow by filename[cr]
* 135bbb = 10111010bbbbbbbb   write to file, byte in low 8 bits of command
* 136000 = 1011110000000000   close write file
* 137000 = 1011111000000000   get usb_error value, follow by read transfer
*
* send one word to interface
S1W    NOP
IF1    OTA USB
IF2    STC USB,C
IF3    SFS USB
       JMP *-1
IF4    CLC USB
       JMP S1W,I
* receive one word from interface
R1W    NOP
IF7    LIA USB
       JMP R1W,I
* command 120xxx - get one byte from VDRIVE2 (bit 15 set if not ready)
G1BS   NOP
       LDA G1BCM
       JSB S1W
       JSB R1W
       JMP G1BS,I
G1BCM  OCT 120000
* command 121bbb - send byte bbb to VDRIVE2
S1BS   NOP
       AND BMASK
       IOR S1BCM
       JSB S1W
       JMP S1BS,I
BMASK  OCT 377
S1BCM  OCT 121000
* command 122xxx - sync with VDRIVE2
SYNC   NOP
       LDA VSCMD
       JSB S1W
       JMP SYNC,I
VSCMD  OCT 122000
* command 123xxx - clear VDRIVE2 responses
CLEAR  NOP
       LDA VCLCM
       JSB S1W
       JMP CLEAR,I
VCLCM  OCT 123000
* command 130xxx - open stream file for read
* must follow by sending filename + cr to interface
ORF    NOP
       LDA ORFCM
       JSB S1W
       JMP ORF,I
ORFCM  OCT 130000
* this sub is not needed for this application...
* streaming is done using PTR-style interface
** command 131xxx - get one byte from input stream file
*GSB    NOP
*       LDA GSBCM
*       JSB S1W
*       JSB R1W
*       JMP GSB,I
*GSBCM  OCT 131000
* command 132xxx - read bytes via PTR interface
RPTR   NOP
       LDA PTRCM
       JSB S1W
       JMP RPTR,I
PTRCM  OCT 132000
* command 134xxx - open stream file for write
* must follow by sending filename + cr to interface
OWF    NOP
       LDA OWFCM
       JSB S1W
       JMP OWF,I
OWFCM  OCT 134000
* command 135bbb - send byte to output stream file
SSB    NOP
       AND BMASK
       IOR SSBCM
       JSB S1W
       JMP SSB,I
SSBCM  OCT 135000
* command 136xxx - close output stream file
CLO    NOP
       LDA CLOCM
       JSB S1W
       JMP CLO,I
CLOCM  OCT 136000
* command 137xxx - get USB status
USTAT  NOP
       LDA USTCM
       JSB S1W
       JSB R1W
       JMP USTAT,I
USTCM  OCT 137000
* console driver (adapted from HP-IPL/OS)
* TTY input
CHRIN   NOP                     * LINKAGE
        LDB CMD2                * GET TTY INPUT COMMAND
TTYP5   OTB TTY                 * SET INPUT MODE
TTYP6   STC TTY,C               * START TTY INTERFACE
TTYP7   SFS TTY                 * GOT CHAR FROM CONSOLE?
        JMP *-1                 * NO, KEEP CHECKING
TTYP8   LIA TTY                 * GET CHAR FROM INTERFACE
TTYPA   CLC TTY                 * TURN OFF INTERFACE
        AND CON2                * AND WITH 7 BIT MASK
        JMP CHRIN,I             * RETURN WITH CHAR IN A
CMD2    OCT 160000              * TTY input command
CON2    OCT 177
* TTY output
CHROT   NOP                     * LINKAGE
        LDB CMD1                * GET TTY OUTPUT COMMAND
TTYP1   OTB TTY                 * SET OUTPUT MODE
        AND CON2                * AND WITH 7 BIT MASK
TTYP2   OTA TTY                 * SEND CHAR TO INTERFACE
TTYP3   STC TTY,C               * SEND CHAR TO CONSOLE
TTYP4   SFS TTY                 * INTERFACE DONE?
        JMP *-1                 * NO, WAIT TILL DONE
TTYP9   CLC TTY                 * TURN OFF INTERFACE
        JMP CHROT,I             * RETURN FROM CHROT
CMD1    OCT 120000              * TTY output command
* mid-level subs...
* initialize VDRIVE2, sync, send ECS[cr]IPA[cr], clear response
VECS   NOP
       JSB SYNC             sync up with VDRIVE
       LDB VINIA            get address of init string
VECS1  LDA 1,I              get doublechar
       SSA                  skip if sign is 0
       JMP VECS2            done sending
       AND HIMSK
       ALF,ALF
       JSB S1BS             send high byte to VDRIVE
       LDA 1,I              get doublechar
       AND BMASK
       JSB S1BS             send low byte to VDRIVE
       INB                  increment pointer
       JMP VECS1            loop to send more bytes
VECS2  JSB CLEAR            clear VDRIVE response
       JMP VECS,I           exit sub
VINIA  DEF *+1              address of chars to send
       OCT 42503,51415,44520,40415    ECS[cr]IPA[cr]
       OCT 177777                     terminate
* get and send line to the VDRIVE2 one character at a time
* no buffering, no backspace/editing - press ESC to abort
* for sending filenames and dos prompt
GSLN   NOP
GSLN1  JSB CHRIN            get character from console
       CPA LFCHR            is it a line feed?
       JMP GSLN1            yes - ignore, get next char
       STA TEMPA            save character to compare
       JSB S1BS             send character to VDRIVE2
       LDA TEMPA            get character that was sent
       CPA RTCHR            was it a return?
       JMP RETEX            yes - exit sub with a line feed
       CPA ESCHR            was it escape?
       JMP ESCEX            yes - escape exit
       JMP GSLN1            no - get more
* exit code... double-check to make sure VDRIVE2 is good with this
ESCEX  JSB SYNC             sync/clear VDRIVE2 error response
RETEX  LDB CRLFA            send space space crlf
       JSB PRTXT
       JMP GSLN,I           exit sub
TEMPA  OCT 0
RTCHR  OCT 15
LFCHR  OCT 12
ESCHR  OCT 33
* get and print VDRIVE2 response
GVDR   NOP
GVDR1  JSB G1BS             get byte from VDRIVE2
       SSA                  skip if bit 15 is not set
       JMP GVDR,I           not ready - exit sub
       CPA LFCHR            is it a line feed?
       RSS                  yes - listing file, skip print
       JSB CHROT            print character
       CPA RTCHR            is it a return character?
       RSS                  yes - skip next
       JMP GVDR1            no - loop for more
       LDA LFCHR            get linefeed char
       JSB CHROT            print it
       JSB DELAY            delay a bit
       JMP GVDR1            loop for more output
* delay sub
DELAY  NOP
       LDB MDELV            get negative delay value
DEL1   JMP *+1              waste time...
       JMP *+1
       JMP *+1
       ISZ 1                increment, skip if zero
       JMP DEL1             not zero - loop
       JMP DELAY,I
MDELV  DEC -20000
* get and send filename to interface
GSFN   NOP
       CLA FNLEN            start with length = 0
GSFN1  JSB CHRIN            get char from user
       STA TEMPA            save to detect escape
       JSB S1W              send to interface
       CPA RTCHR            was enter pressed
       JMP GSFNR            yes - exit sub with a line feed
       CPA ESCHR            was escape pressed
       JMP GSFNE            yes - jump to escape exit
       ISZ FNLEN            no - increment filename length
       JMP GSFN1            loop for more chars
GSFNE  LDA RTCHR            escape exit - send CRLF to console
       JSB S1W              send CR to interface
       JSB CLEAR            clear VDRIVE2 error response
GSFNR  LDB CRLFA
       JSB PRTXT            print space space crlf
       JMP GSFN,I
FNLEN  OCT 0                length of filename
* open read filename
ORFN   NOP
       JSB ORF              send open read file command
       JSB GSFN             get and send filename
       JMP ORFN,I           exit sub
* open write filename
OWFN   NOP
       JSB OWF              send open write file command
       JSB GSFN             get and send filename
       JMP OWFN,I           exit sub
* "dos" prompt... press ESC to exit 
DOSPR  NOP
       JSB VECS             initialize VDRIVE2 for "human"
DOS1   JSB GSLN             get chars and send to VDRIVE2
       LDA TEMPA            get last char pressed
       CPA ESCHR            was it escape?
       JMP DOSPR,I          yes - exit sub
       JSB DELAY            delay a bit
       JSB GVDR             print response
       JMP DOS1             get next dos command
* attach file to PTR
ATPTR  NOP
       JSB ORFN             open read filename
       LDA TEMPA            get last char pressed
       CPA ESCHR            was it escape?
       JMP ATPTR,I          yes - exit sub
       JSB USTAT            get USB status
       OTA 1                display status
       SZA,RSS              skip if error
       JSB RPTR             enable PTR mode if no error
       JMP ATPTR,I          exit sub
* save binary file in ABS format
* saves from 2-17677 (all memory except for boot loader)
SVBIN  NOP
* detect whether PTP or VDRIVE2 firmware...
       JSB CLO              close write file
       CLA
       JSB SSB              try to write a 0 byte
       JSB USTAT            get status - should be error
       SZA,RSS              if error skip to get filename
       JMP DOPTP            no error returned, do a PTP-style save
       JSB OWFN             open write filename
       LDA FNLEN            get length of filename typed
       SZA,RSS              skip if non-zero
       JMP SVBIN,I          exit sub if nothing entered
       LDA TEMPA            get last char
       CPA ESCHR            esc?
       JMP SVBIN,I          yes - exit sub
       JSB USTAT            get USB status
       OTA 1                display status
       SZA                  skip if no error
       JMP SVBIN,I          exit if error
DOSAV  JSB SZERB            send 16 zero bytes (leader)
       LDA STADR
       STA MEMCT            set starting address
* encode loop
SVB1   LDA ENADR
       LDB MEMCT
       CMB,INB
       ADA 1                A = end - counter
       STA REMWD            save remaining words to do
       SZA,RSS
       JMP SVBEN            if 0 then at end, finish write
       LDB RECSZ
       CMB,INB
       ADA 1                A = remaining - record size
       LDB RECSZ            default record size
       SSA                  skip if complete record
       LDB REMWD            number of words in last record
       STB RECLM            save for record limit
       LDA 1                put in A
       JSB SSB              write # words in record
       CLA
       JSB SSB              write a zero byte
       LDA MEMCT            get mem address
       STA CHKSM            store in checksum
       AND HIMSK
       ALF,ALF
       JSB SSB              send high byte of address
       LDA MEMCT
       AND BMASK
       JSB SSB              send low byte of address
       CLA
       STA RECCT            clear record counter
* word loop
SVB2   LDA MEMCT,I          get word from memory
       LDB CHKSM
       ADB 0
       STB CHKSM            add to checksum
       STA 1                save mem word to buffer
       AND HIMSK
       ALF,ALF
       JSB SSB              send high byte of word
       LDA 1                get buffered memory word
       AND BMASK
       JSB SSB              send low byte of word
       ISZ MEMCT            increment memory counter
       ISZ RECCT            increment record counter
       LDA RECCT
       CPA RECLM            done?
       RSS                  yes - skip to write checksum
       JMP SVB2             no - write more words
       LDA CHKSM
       AND HIMSK
       ALF,ALF
       JSB SSB              write high byte of checksum
       LDA CHKSM
       AND BMASK
       JSB SSB              write low byte of checksum
       JMP SVB1             loop to write another record
* finish writing binary
SVBEN  JSB SZERB            send 16 zero bytes (trailer)
       JSB CLO              close write file
       JSB USTAT            get status
       OTA 1                display status
       JMP SVBIN,I          exit sub
* do a PTP save - this works because the VDRIVE interface firmware is
* very simple, disk commands have low byte clear so just adds leader
DOPTP  LDB PTPTA
       JSB PRTXT
       JMP DOSAV
PTPTA  DEF *+1
       ASC 3,[PTP]          tell user it's doing a PTP save
       OCT 6412,177777
STADR  OCT 2                start address of save
ENADR  DEF TOMEM            end address + 1 of save
MEMCT  OCT 0                mem counter
RECCT  OCT 0                record counter
RECLM  OCT 0                words in current record
CHKSM  OCT 0                record checksum
* RECSZ  OCT 33               record size (words in each record)
RECSZ  EQU ESCHR            save a word
REMWD  OCT 0                remaining words
HIMSK  OCT 177400           high-byte mask (reuse BMASK for low mask)
* sub to write leader/trailer
SZERB  NOP
       LDB NZERC            get negative zero count
SZER1  CLA
       JSB SSB              send zero byte
       ISZ 1                inc counter, skip if done
       JMP SZER1            send more zero bytes
       JMP SZERB,I          exit sub
NZERC  DEC -16              send 16 zero bytes
* print subroutine, address in B
* terminates if bit 15 of text word set
PRPTR  OCT 0                print pointer
PRTXT  NOP
       STB PRPTR
PRTX1  LDA PRPTR,I          get packed double char
       SSA                  skip if sign is 0
       JMP PRTXT,I          exit sub if sign set
       AND HIMSK
       ALF,ALF
       JSB CHROT            print high byte
       LDA PRPTR,I
       AND BMASK
       SZA                  skip if zero
       JSB CHROT            print low byte
       ISZ PRPTR            inc print pointer
       JMP PRTX1            loop to print more
* main menu
MENU   LDB MENPR
       JSB PRTXT            print menu
MENU1  JSB CHRIN            get selection
       CPA OPT1             is it 1) save binary?
       JMP OPT1C            yes - do it
       CPA OPT2             is it 2) attach file?
       JMP OPT2C            yes - do it
       CPA OPT3             is it 3) dos prompt?
       JMP OPT3C            yes - do it
       CPA OPT4             is it 4) run from 2
       JMP OPT4C            yes - do it
       JMP MENU1            no - get another keystroke
OPT1   DEC 49
OPT1C  LDB SVPRM
       JSB PRTXT            print "SAVE TO: "
       JSB SVBIN
       JMP MENU
OPT2   DEC 50
OPT2C  LDB ATPRM
       JSB PRTXT            print "ATTACH FROM: "
       JSB ATPTR
       JMP MENU
OPT3   DEC 51
OPT3C  LDB PRINS
       JSB PRTXT            print dos prompt instructions
       JSB DOSPR
       JMP MENU
OPT4   DEC 52
OPT4C  LDB CRLFA
       JSB PRTXT
       JMP *+1,I
       OCT 2
MENPR  DEF *+1
       OCT 6412
       ASC 7,1) SAVE BINARY
       OCT 6412
       ASC 7,2) ATTACH FILE
       OCT 6412
       ASC 7,3) DOS PROMPT
       OCT 6412
       ASC 7,4) RUN FROM 2
       OCT 6412
       ASC 1,>  
       OCT 177777
SVPRM  DEF *+1
       OCT 6412
       ASC 5,SAVE TO: 
       OCT 177777
ATPRM  DEF *+1
       OCT 6412
       ASC 7,ATTACH FROM: 
       OCT 177777
PRINS  DEF *+1
       OCT 6412
       ASC 6,ESC TO EXIT
CRLFP  OCT 6412
       OCT 20015
       OCT 177777
CRLFA  DEF CRLFP
END
